import base64
import random
import string

import webauthn
from django.contrib.sites.shortcuts import get_current_site

from . import app_settings
from .models import WebauthnData


def user_has_webauthn_enabled(user):
    if not user.is_authenticated:  # pragma: no cover
        return False
    return WebauthnData.objects.filter(user=user).exists()


def random_numbers_letters(count):  # pragma: no cover
    return "".join([random.SystemRandom().choice(string.ascii_letters + string.digits) for i in range(count)])


def get_icon_url(request):
    if app_settings.ICON_URL is None:  # pragma: no cover
        return f"https://{get_site_domain(request)}/favicon.ico"
    return app_settings.ICON_URL


def get_site_domain(request):
    if app_settings.DOMAIN is None:  # pragma: no cover
        site = get_current_site(request)
        return site.domain
    return app_settings.DOMAIN


def get_origin(request):
    if app_settings.ORIGIN is None:  # pragma: no cover
        return f"https://{get_site_domain(request)}"
    return app_settings.ORIGIN


def get_site_name(request):
    if app_settings.NAME is None:  # pragma: no cover
        site = get_current_site(request)
        return site.name
    return app_settings.NAME


def get_display_name(request):
    return (f"{get_site_name(request)} user: {request.user.get_username()}",)


def sanitize_session(request):
    """
    If present, remove artifact from previous aborted login, or successful login.
    """
    try:
        del request.session["allauth_webauthn_user_id"]
    except KeyError:
        pass
    try:
        del request.session["allauth_webauthn_challenge"]
    except KeyError:
        pass


def authenticate(request, user_id, credential_id, data):
    """Authenticate a dedicated user given a signed token."""
    if not user_id or not credential_id:  # pragma: no cover
        return None

    challenge = request.session.get("allauth_webauthn_challenge")
    if not challenge:  # pragma: no cover
        return None

    device = WebauthnData.objects.filter(user_id=user_id, credential_id=credential_id).first()
    if not device:
        return None

    webauthn_user = webauthn.WebAuthnUser(
        user_id=base64.b64encode(str(device.user.id).encode()).decode(),
        username=device.user.get_username(),
        display_name=get_display_name(request),
        icon_url=get_icon_url(request),
        credential_id=device.credential_id,
        public_key=device.public_key,
        sign_count=device.sign_counter,
        rp_id=get_site_domain(request),
    )

    webauthn_assertion_response = webauthn.WebAuthnAssertionResponse(
        webauthn_user, data, challenge, get_origin(request), uv_required=False
    )

    try:
        sign_counter = webauthn_assertion_response.verify()
    except Exception:
        return None

    device.update_sign_counter(sign_counter)

    return device.user



from django.conf import settings
from django.db import models
from django.utils.timezone import now


class WebauthnData(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=128, blank=True)
    credential_id = models.CharField(max_length=250, unique=True)
    public_key = models.CharField(max_length=65)
    sign_counter = models.IntegerField(default=0)
    last_used_on = models.DateTimeField(default=now)

    def update_sign_counter(self, sign_counter):
        """Updates the sign_counter and last_used_on."""
        self.sign_counter = sign_counter
        self.last_used_on = now()
        self.save()

    def __str__(self):
        return self.name or str(self.user)



import base64

import webauthn
from allauth.account import signals
from allauth.account.adapter import get_adapter
from allauth.account.utils import get_login_redirect_url
from allauth.account.utils import get_next_redirect_url
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.http.response import JsonResponse
from django.shortcuts import redirect
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from django.views.generic.base import TemplateView

from . import app_settings
from .models import WebauthnData
from .utils import authenticate
from .utils import get_display_name
from .utils import get_icon_url
from .utils import get_origin
from .utils import get_site_domain
from .utils import get_site_name
from .utils import random_numbers_letters
from .utils import sanitize_session


class Register(LoginRequiredMixin, View):
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        site_name = get_site_name(request)
        challenge = random_numbers_letters(32)
        request.session["allauth_webauthn_challenge"] = challenge
        reg_data = webauthn.WebAuthnMakeCredentialOptions(
            challenge=challenge,
            rp_name=site_name,
            rp_id=get_site_domain(request),
            user_id=base64.b64encode(str(request.user.id).encode()).decode(),
            username=request.user.get_username(),
            display_name=get_display_name(request),
            icon_url=get_icon_url(request),
        ).registration_dict
        return JsonResponse(reg_data)

    def post(self, request, *args, **kwargs):
        site_domain = get_site_domain(request)
        challenge = request.session.get("allauth_webauthn_challenge")
        if not challenge:
            return HttpResponse("No challenge exists in your session.", status=422)
        registration_response = request.POST
        webauthn_registration_response = webauthn.WebAuthnRegistrationResponse(
            rp_id=site_domain,
            origin=get_origin(request),
            registration_response=registration_response,
            challenge=challenge,
            self_attestation_permitted=True,
            none_attestation_permitted=True,
            uv_required=False,
        )
        try:
            webauthn_credential = webauthn_registration_response.verify()
        except Exception as e:
            messages.error(request, _("Registration failed. Error: %(error)s") % {"error": e})
            return redirect(app_settings.REGISTRATION_ERROR_URL)

        credential_id = str(webauthn_credential.credential_id, "utf-8")
        public_key = str(webauthn_credential.public_key, "utf-8")
        sign_count = webauthn_credential.sign_count

        device = WebauthnData.objects.filter(credential_id=credential_id)
        if device.exists():
            messages.error(
                request,
                _("This token is already registered to an account. Try logging in with it."),
            )
            return redirect(app_settings.REGISTRATION_ERROR_URL)

        token_count = WebauthnData.objects.filter(user=request.user).count()
        device_name = _("Device #%(num)d") % {"num": token_count + 1}
        WebauthnData.objects.create(
            user=request.user,
            name=device_name,
            credential_id=credential_id,
            public_key=public_key,
            sign_counter=sign_count,
        )

        messages.success(request, _("Your security token has been successfully registered."))
        return redirect(app_settings.REGISTRATION_REDIRECT_URL)


class Login(TemplateView):
    def get_template_names(self):
        return [app_settings.get_login_template()]

    def dispatch(self, request, *args, **kwargs):
        # Redirect the user to the login page if they does not come from there,
        # i.e. if "allauth_webauthn_user_id" is not included in the session.
        if "allauth_webauthn_user_id" not in request.session:
            return redirect("account_login")
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        kwargs["login_redirect_url"] = get_next_redirect_url(self.request)
        return super().get(request, *args, **kwargs)


class Verify(View):
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        site_domain = get_site_domain(request)
        challenge = random_numbers_letters(32)
        request.session["allauth_webauthn_challenge"] = challenge
        user_id = request.session.get("allauth_webauthn_user_id")
        user_credential_ids = (
            WebauthnData.objects.filter(user_id=user_id)
            .order_by("-last_used_on")
            .values_list("credential_id", flat=True)
        )
        login_data = {
            "challenge": challenge,
            "timeout": 60000,
            "rpId": site_domain,
            "allowCredentials": [{"id": credential_id, "type": "public-key"} for credential_id in user_credential_ids],
            "userVerification": "preferred",
        }
        return JsonResponse(login_data)

    def post(self, request, *args, **kwargs):
        user_id = request.session.get("allauth_webauthn_user_id")
        challenge = request.session.get("allauth_webauthn_challenge")
        if not challenge or not user_id:
            messages.error(request, "No challenge or user exists for your session.")
            return redirect(app_settings.LOGIN_ERROR_URL)

        user = authenticate(request, user_id, request.POST["id"], request.POST)
        if user is None:
            messages.error(request, "Your credentials could not be validated.")
            return redirect(app_settings.LOGIN_ERROR_URL)

        adapter = get_adapter(request)

        adapter.login_without_webauthn(request, user)

        # Perform the rest of allauth.account.utils.perform_login, this is
        # copied from commit cedad9f156a8c78bfbe43a0b3a723c1a0b840dbd.

        # TODO Support redirect_url.
        response = HttpResponseRedirect(get_login_redirect_url(self.request))

        # TODO Support signal_kwargs.
        signals.user_logged_in.send(sender=user.__class__, request=self.request, response=response, user=user)

        adapter.add_message(
            self.request,
            messages.SUCCESS,
            "account/messages/logged_in.txt",
            {"user": user},
        )

        sanitize_session(request)

        return response


class Remove(LoginRequiredMixin, View):
    def post(self, request, *args, **kwargs):
        device = WebauthnData.objects.get(user=request.user, pk=kwargs["pk"])
        device.delete()
        return redirect(app_settings.REMOVE_RENAME_REDIRECT_URL)


class Rename(LoginRequiredMixin, View):
    def post(self, request, *args, **kwargs):
        device = WebauthnData.objects.get(user=request.user, pk=kwargs["pk"])
        device.name = request.POST["name"]
        device.save()
        return redirect(app_settings.REMOVE_RENAME_REDIRECT_URL)



from django.urls import path

from . import views

urlpatterns = [
    path("login/", views.Login.as_view(), name="webauthn-login"),
    path("register/", views.Register.as_view(), name="webauthn-register"),
    path("verify/", views.Verify.as_view(), name="webauthn-verify"),
    path("remove/<int:pk>/", views.Remove.as_view(), name="webauthn-remove"),
    path("rename/<int:pk>/", views.Rename.as_view(), name="webauthn-rename"),
]



from django.conf import settings

ICON_URL = getattr(settings, "DJANGO_ALLAUTH_WEBAUTHN_ICON_URL", None)
DOMAIN = getattr(settings, "DJANGO_ALLAUTH_WEBAUTHN_DOMAIN", None)
NAME = getattr(settings, "DJANGO_ALLAUTH_WEBAUTHN_NAME", None)
ORIGIN = getattr(settings, "DJANGO_ALLAUTH_WEBAUTHN_ORIGIN", None)
# Where to redirect after successful key registration.
REGISTRATION_REDIRECT_URL = getattr(
    settings,
    "DJANGO_ALLAUTH_WEBAUTHN_REGISTRATION_REDIRECT_URL",
    settings.LOGIN_REDIRECT_URL,
)
# Where to redirect after key registration error.
REGISTRATION_ERROR_URL = getattr(
    settings,
    "DJANGO_ALLAUTH_WEBAUTHN_REGISTRATION_ERROR_URL",
    settings.LOGIN_REDIRECT_URL,
)
# Where to redirect after login errors.
LOGIN_ERROR_URL = getattr(settings, "DJANGO_ALLAUTH_WEBAUTHN_LOGIN_ERROR_URL", settings.LOGIN_URL)
# Where to redirect after remove or renaming a token?
REMOVE_RENAME_REDIRECT_URL = getattr(
    settings,
    "DJANGO_ALLAUTH_WEBAUTHN_REMOVE_RENAME_REDIRECT_URL",
    settings.LOGIN_REDIRECT_URL,
)


def get_login_template():
    """returns the login template"""
    return getattr(settings, "DJANGO_ALLAUTH_WEBAUTHN_LOGIN_TEMPLATE", "django_allauth_webauthn/login.html")



# Generated by Django 3.2.7 on 2021-09-14 19:31
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="WebauthnData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(blank=True, max_length=128)),
                ("credential_id", models.CharField(max_length=250, unique=True)),
                ("public_key", models.CharField(max_length=65)),
                ("sign_counter", models.IntegerField(default=0)),
                (
                    "last_used_on",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from django.contrib import admin

from .models import WebauthnData


@admin.register(WebauthnData)
class WebauthnDataAdmin(admin.ModelAdmin):
    search_fields = ["name", "credential_id"]
    list_display = ["user", "name", "last_used_on"]



from urllib.parse import urlencode

from allauth.account.adapter import DefaultAccountAdapter
from allauth.exceptions import ImmediateHttpResponse
from django.http import HttpResponseRedirect
from django.urls import reverse

from .utils import sanitize_session
from .utils import user_has_webauthn_enabled


class WebAuthnAdapter(DefaultAccountAdapter):
    def has_webauthn_enabled(self, user):
        """Returns True if the user has to user webauthn as a 2nd login factor."""
        return user_has_webauthn_enabled(user)

    def login(self, request, user):
        # Clean previous logins or failed logins (e.g. aborted during webauth)
        sanitize_session(request)
        # Require two-factor authentication if it has been configured.
        if self.has_webauthn_enabled(user):
            # Cast to string for the case when this is not a JSON serializable
            # object, e.g. a UUID.
            request.session["allauth_webauthn_user_id"] = str(user.id)

            redirect_url = reverse("webauthn-login")
            # Add "next" parameter to the URL.
            view = request.resolver_match.func.view_class()
            view.request = request
            success_url = view.get_success_url()
            query_params = request.GET.copy()
            if success_url:
                query_params[view.redirect_field_name] = success_url
            if query_params:
                redirect_url += "?" + urlencode(query_params)

            raise ImmediateHttpResponse(response=HttpResponseRedirect(redirect_url))

        # Otherwise defer to the original allauth adapter.
        return self.login_without_webauthn(request, user)

    def login_without_webauthn(self, request, user):
        """Perform a login with the original allauth adapter"""
        return super().login(request, user)
